from composeml.data_slice.extension import DataSliceContext, DataSliceFrame


class DataSliceGenerator:
    """Generates data slices for the lable maker."""

    def __init__(
        self, window_size, gap=None, min_data=None, max_data=None, drop_empty=True
    ):
        self.window_size = window_size
        self.gap = gap
        self.min_data = min_data
        self.max_data = max_data
        self.drop_empty = drop_empty

    def __call__(self, df):
        """Applies the data slice generator to the data frame."""
        is_column = self.window_size in df
        method = "column" if is_column else "time"
        attr = f"_slice_by_{method}"
        return getattr(self, attr)(df)

    def _slice_by_column(self, df):
        """Slices the data frame by an existing column."""
        slices = df.groupby(self.window_size, sort=False)
        slice_number = 1

        for group, ds in slices:
            ds = DataSliceFrame(ds)
            ds.context = DataSliceContext(
                slice_number=slice_number,
                slice_start=ds.first_valid_index(),
                slice_stop=ds.last_valid_index(),
            )
            setattr(ds.context, self.window_size, group)
            del ds.context.next_start
            slice_number += 1
            yield ds

    def _slice_by_time(self, df):
        """Slices the data frame along the time index."""
        data_slices = df.slice(
            size=self.window_size,
            start=self.min_data,
            stop=self.max_data,
            step=self.gap,
            drop_empty=self.drop_empty,
        )

        for ds in data_slices:
            yield ds



import pandas as pd
from pytest import raises


def test_bins(labels):
    given_labels = labels.bin(2)
    transform = given_labels.transforms[0]

    assert transform["transform"] == "bin"
    assert transform["bins"] == 2
    assert transform["quantiles"] is False
    assert transform["labels"] is None
    assert transform["right"] is True

    answer = [
        pd.Interval(157.5, 283.46, closed="right"),
        pd.Interval(31.288, 157.5, closed="right"),
        pd.Interval(157.5, 283.46, closed="right"),
        pd.Interval(31.288, 157.5, closed="right"),
    ]

    answer = pd.Categorical(answer, ordered=True)
    labels = labels.assign(my_labeling_function=answer)
    pd.testing.assert_frame_equal(given_labels, labels)


def test_quantile_bins(labels):
    given_labels = labels.bin(2, quantiles=True)
    transform = given_labels.transforms[0]

    assert transform["transform"] == "bin"
    assert transform["bins"] == 2
    assert transform["quantiles"] is True
    assert transform["labels"] is None
    assert transform["right"] is True

    answer = [
        pd.Interval(137.44, 283.46, closed="right"),
        pd.Interval(31.538999999999998, 137.44, closed="right"),
        pd.Interval(137.44, 283.46, closed="right"),
        pd.Interval(31.538999999999998, 137.44, closed="right"),
    ]

    answer = pd.Categorical(answer, ordered=True)
    labels = labels.assign(my_labeling_function=answer)
    pd.testing.assert_frame_equal(given_labels, labels)


def test_single_target(total_spent):
    lt = total_spent.copy()
    lt.target_columns.append("target_2")
    match = "must first select an individual target"
    with raises(AssertionError, match=match):
        lt.bin(2)



import os

import pandas as pd

DATA = os.path.join(os.path.dirname(__file__))


def load_transactions():
    path = os.path.join(DATA, "transactions.csv")
    df = pd.read_csv(path, parse_dates=["transaction_time"])
    return df



import pandas as pd


def test_lead(labels):
    labels = labels.apply_lead("10min")
    transform = labels.transforms[0]

    assert transform["transform"] == "apply_lead"
    assert transform["value"] == "10min"

    answer = [
        "2014-01-01 00:35:00",
        "2014-01-01 00:38:00",
        "2013-12-31 23:51:00",
        "2013-12-31 23:54:00",
    ]

    time = pd.Series(answer, name="time", dtype="datetime64[ns]")
    time = time.rename_axis("label_id")

    pd.testing.assert_series_equal(labels["time"], time)



import pandas as pd


def describe_label_times(label_times):
    """Prints out label info with transform settings that reproduce labels."""
    target_column = label_times.target_columns[0]
    is_discrete = label_times.is_discrete[target_column]

    if is_discrete:
        distribution = label_times[target_column].value_counts()
        distribution.sort_index(inplace=True)
        distribution.index = distribution.index.astype("str")
        distribution["Total:"] = distribution.sum()
    else:
        distribution = label_times[target_column].describe()

    print("Label Distribution\n" + "-" * 18, end="\n")
    print(distribution.to_string(), end="\n\n\n")

    metadata = label_times.settings
    target_column = metadata["label_times"]["target_columns"][0]
    target_type = metadata["label_times"]["target_types"][target_column]
    target_dataframe_index = metadata["label_times"]["target_dataframe_index"]

    settings = {
        "target_column": target_column,
        "target_dataframe_index": target_dataframe_index,
        "target_type": target_type,
    }

    settings.update(metadata["label_times"]["search_settings"])
    settings = pd.Series(settings)

    print("Settings\n" + "-" * 8, end="\n")
    settings.sort_index(inplace=True)
    print(settings.to_string(), end="\n\n\n")

    print("Transforms\n" + "-" * 10, end="\n")
    transforms = metadata["label_times"]["transforms"]
    for step, transform in enumerate(transforms):
        transform = pd.Series(transform)
        transform.sort_index(inplace=True)
        name = transform.pop("transform")
        transform = transform.add_prefix("  - ")
        transform = transform.add_suffix(":")
        transform = transform.to_string()
        header = "{}. {}\n".format(step + 1, name)
        print(header + transform, end="\n\n")

    if len(transforms) == 0:
        print("No transforms applied", end="\n\n")



import pandas as pd
import pytest

from composeml import LabelMaker
from composeml.tests.utils import to_csv


def test_search_default(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
    )

    given_labels = lm.search(transactions, num_examples_per_instance=1)
    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,2",
        "1,2019-01-01 09:00:00,3",
        "2,2019-01-01 10:30:00,4",
        "3,2019-01-01 12:30:00,1",
    ]

    assert given_labels == labels


def test_search_examples_per_label(transactions, total_spent_fn):
    def total_spent(ds):
        return total_spent_fn(ds) > 2

    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent,
    )

    n_examples = {True: -1, False: 1}
    given_labels = lm.search(transactions, num_examples_per_instance=n_examples, gap=1)
    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,False",
        "1,2019-01-01 09:00:00,True",
        "1,2019-01-01 09:30:00,False",
        "2,2019-01-01 10:30:00,True",
        "2,2019-01-01 11:00:00,True",
        "2,2019-01-01 11:30:00,False",
        "3,2019-01-01 12:30:00,False",
    ]

    assert given_labels == labels


def test_search_with_undefined_labels(transactions, total_spent_fn):
    def total_spent(ds):
        return total_spent_fn(ds) % 3

    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent,
    )

    n_examples = {1: 1, 2: 1}
    given_labels = lm.search(transactions, num_examples_per_instance=n_examples, gap=1)
    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,2",
        "0,2019-01-01 08:30:00,1",
        "1,2019-01-01 09:30:00,2",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 10:30:00,1",
        "2,2019-01-01 11:30:00,2",
        "3,2019-01-01 12:30:00,1",
    ]

    assert given_labels == labels


def test_search_with_multiple_targets(transactions, total_spent_fn, unique_amounts_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        window_size=2,
        labeling_function={
            "total_spent": total_spent_fn,
            "unique_amounts": unique_amounts_fn,
        },
    )

    expected = [
        "customer_id,time,total_spent,unique_amounts",
        "0,2019-01-01 08:00:00,2,1",
        "1,2019-01-01 09:00:00,2,1",
        "1,2019-01-01 10:00:00,1,1",
        "2,2019-01-01 10:30:00,2,1",
        "2,2019-01-01 11:30:00,2,1",
        "3,2019-01-01 12:30:00,1,1",
    ]

    lt = lm.search(transactions, num_examples_per_instance=-1)
    actual = lt.pipe(to_csv, index=False)
    info = "unexpected calculated values"
    assert actual == expected, info

    expected = [
        "customer_id,time,unique_amounts",
        "0,2019-01-01 08:00:00,1",
        "1,2019-01-01 09:00:00,1",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 10:30:00,1",
        "2,2019-01-01 11:30:00,1",
        "3,2019-01-01 12:30:00,1",
    ]

    actual = lt.select("unique_amounts")
    actual = actual.pipe(to_csv, index=False)
    info = "selected values differ from calculated values"
    assert actual == expected, info


def test_search_offset_mix_0(transactions, total_spent_fn):
    """
    Test offset mix with window_size (absolute), minimum_data (absolute), and gap (absolute).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size="2h",
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=2,
        minimum_data="30min",
        gap="2h",
        drop_empty=True,
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:30:00,1",
        "1,2019-01-01 09:30:00,2",
        "2,2019-01-01 11:00:00,3",
    ]

    assert given_labels == labels


def test_search_offset_mix_1(transactions, total_spent_fn):
    """
    Test offset mix with window_size (relative), minimum_data (absolute), and gap (absolute).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size=4,
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=2,
        minimum_data="2019-01-01 10:00:00",
        gap="4h",
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 10:00:00,4",
        "3,2019-01-01 10:00:00,1",
    ]

    assert given_labels == labels


def test_search_offset_mix_2(transactions, total_spent_fn):
    """
    Test offset mix with window_size (absolute), minimum_data (relative), and gap (absolute).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size="30min",
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=2,
        minimum_data=2,
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 11:30:00,1",
        "2,2019-01-01 12:00:00,1",
    ]

    assert given_labels == labels


def test_search_offset_mix_3(transactions, total_spent_fn):
    """
    Test offset mix with window_size (absolute), minimum_data (absolute), and gap (relative).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size="8h",
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=-1,
        minimum_data="2019-01-01 08:00:00",
        gap=1,
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,2",
        "0,2019-01-01 08:30:00,1",
        "1,2019-01-01 09:00:00,3",
        "1,2019-01-01 09:30:00,2",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 10:30:00,4",
        "2,2019-01-01 11:00:00,3",
        "2,2019-01-01 11:30:00,2",
        "2,2019-01-01 12:00:00,1",
        "3,2019-01-01 12:30:00,1",
    ]

    assert given_labels == labels


def test_search_offset_mix_4(transactions, total_spent_fn):
    """
    Test offset mix with window_size (relative), minimum_data (relative), and gap (absolute).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size=1,
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=2,
        gap="30min",
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,1",
        "0,2019-01-01 08:30:00,1",
        "1,2019-01-01 09:00:00,1",
        "1,2019-01-01 09:30:00,1",
        "2,2019-01-01 10:30:00,1",
        "2,2019-01-01 11:00:00,1",
        "3,2019-01-01 12:30:00,1",
    ]

    assert given_labels == labels


def test_search_offset_mix_5(transactions, total_spent_fn):
    """
    Test offset mix with window_size (relative), minimum_data (absolute), and gap (relative).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size=2,
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=2,
        minimum_data="1h",
        gap=2,
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 11:30:00,2",
    ]

    assert given_labels == labels


def test_search_offset_mix_6(transactions, total_spent_fn):
    """
    Test offset mix with window_size (absolute), minimum_data (relative), and gap (relative).
    """
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size="1h",
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=1,
        minimum_data=3,
        gap=1,
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "2,2019-01-01 12:00:00,1",
    ]

    assert given_labels == labels


def test_search_offset_mix_7(transactions, total_spent_fn):
    """
    Test offset mix with window_size (relative), minimum_data (relative), and gap (relative).
    """

    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size=10,
    )

    given_labels = lm.search(
        transactions,
        num_examples_per_instance=float("inf"),
    )

    given_labels = to_csv(given_labels, index=False)

    labels = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,2",
        "1,2019-01-01 09:00:00,3",
        "2,2019-01-01 10:30:00,4",
        "3,2019-01-01 12:30:00,1",
    ]

    assert given_labels == labels


def test_search_offset_negative_0(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=lambda: None,
        window_size=2,
    )

    match = "offset must be positive"
    with pytest.raises(AssertionError, match=match):
        lm.search(
            transactions,
            num_examples_per_instance=2,
            minimum_data=-1,
            gap=-1,
        )


def test_search_offset_negative_1(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=lambda: None,
        window_size=2,
    )

    match = "offset must be positive"
    with pytest.raises(AssertionError, match=match):
        lm.search(
            transactions,
            num_examples_per_instance=2,
            minimum_data="-1h",
            gap="-1h",
        )


def test_search_invalid_n_examples(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
    )

    with pytest.raises(AssertionError, match="must specify gap"):
        next(lm.slice(transactions, num_examples_per_instance=2))

    with pytest.raises(AssertionError, match="must specify gap"):
        lm.search(transactions, num_examples_per_instance=2)


def test_column_based_windows(transactions, total_spent_fn):
    session_id = [1, 2, 3, 3, 4, 5, 5, 5, 6, 7]
    df = transactions.assign(session_id=session_id)

    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        window_size="session_id",
        labeling_function=total_spent_fn,
    )

    actual = lm.search(df, -1).pipe(to_csv, index=False)

    expected = [
        "customer_id,time,total_spent",
        "0,2019-01-01 08:00:00,1",
        "0,2019-01-01 08:30:00,1",
        "1,2019-01-01 09:00:00,2",
        "1,2019-01-01 10:00:00,1",
        "2,2019-01-01 10:30:00,3",
        "2,2019-01-01 12:00:00,1",
        "3,2019-01-01 12:30:00,1",
    ]

    assert actual == expected


def test_search_with_invalid_index(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=lambda df: None,
        window_size=2,
    )

    df = transactions.sample(n=10, random_state=0)
    match = "data frame must be sorted chronologically"
    with pytest.raises(AssertionError, match=match):
        lm.search(df, num_examples_per_instance=2)

    df = transactions.assign(time=pd.NaT)
    match = "index contains null values"
    with pytest.raises(AssertionError, match=match):
        lm.search(df, num_examples_per_instance=2)


def test_search_on_empty_labels(transactions):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=lambda ds: None,
        window_size=2,
    )

    given_labels = lm.search(
        transactions,
        minimum_data=1,
        num_examples_per_instance=2,
        gap=1,
    )

    assert given_labels.empty


def test_data_slice_overlap(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
        window_size="1h",
    )

    for ds in lm.slice(transactions, num_examples_per_instance=2):
        overlap = ds.index == ds.context.slice_stop
        assert not overlap.any()


def test_label_type(transactions, total_spent_fn):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=total_spent_fn,
    )
    lt = lm.search(transactions, num_examples_per_instance=1)
    assert lt.target_types["total_spent"] == "continuous"
    assert lt.bin(2).target_types["total_spent"] == "discrete"


def test_search_with_maximum_data(transactions):
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        labeling_function=len,
        window_size="1h",
    )

    lt = lm.search(
        df=transactions.sort_values("time"),
        num_examples_per_instance=-1,
        minimum_data="2019-01-01 08:00:00",
        maximum_data="2019-01-01 09:00:00",
        drop_empty=False,
    )

    expected = [
        "customer_id,time,len",
        "0,2019-01-01 08:00:00,2",
        "0,2019-01-01 09:00:00,0",
        "1,2019-01-01 08:00:00,0",
        "1,2019-01-01 09:00:00,2",
        "2,2019-01-01 08:00:00,0",
        "2,2019-01-01 09:00:00,0",
        "3,2019-01-01 08:00:00,0",
        "3,2019-01-01 09:00:00,0",
    ]

    actual = lt.pipe(to_csv, index=False)
    assert actual == expected

    lt = lm.search(
        df=transactions.sort_values("time"),
        num_examples_per_instance=-1,
        maximum_data="30min",
        drop_empty=False,
        gap="30min",
    )

    expected = [
        "customer_id,time,len",
        "0,2019-01-01 08:00:00,2",
        "0,2019-01-01 08:30:00,1",
        "1,2019-01-01 09:00:00,2",
        "1,2019-01-01 09:30:00,2",
        "2,2019-01-01 10:30:00,2",
        "2,2019-01-01 11:00:00,2",
        "3,2019-01-01 12:30:00,1",
        "3,2019-01-01 13:00:00,0",
    ]

    actual = lt.pipe(to_csv, index=False)
    assert actual == expected


@pytest.mark.parametrize(
    "minimum_data",
    [
        {1: "2019-01-01 09:30:00", 2: "2019-01-01 11:30:00"},
        {1: "30min", 2: "1h"},
        {1: 1, 2: 2},
    ],
)
def test_minimum_data_per_group(transactions, minimum_data):
    lm = LabelMaker(
        "customer_id", labeling_function=len, time_index="time", window_size="1h"
    )
    for supported_type in [minimum_data, pd.Series(minimum_data)]:
        lt = lm.search(transactions, 1, minimum_data=supported_type)
        actual = to_csv(lt, index=False)

        expected = [
            "customer_id,time,len",
            "1,2019-01-01 09:30:00,2",
            "2,2019-01-01 11:30:00,2",
        ]

        assert actual == expected


def test_minimum_data_per_group_error(transactions):
    lm = LabelMaker(
        "customer_id", labeling_function=len, time_index="time", window_size="1h"
    )
    data = ["2019-01-01 09:00:00", "2019-01-01 12:00:00"]
    minimum_data = pd.Series(data=data, index=[1, 1])
    match = "more than one cutoff time exists for a target group"

    with pytest.raises(ValueError, match=match):
        lm.search(transactions, 1, minimum_data=minimum_data)


def test_label_maker_categorical_target_with_missing_data(transactions, total_spent_fn):
    transactions = transactions.copy()
    transactions["customer_id"] = transactions["customer_id"].astype("category")
    lm = LabelMaker(
        target_dataframe_index="customer_id",
        time_index="time",
        window_size=3,
        labeling_function=total_spent_fn,
    )
    # use on only the first 8 rows so the df will not contain data for customer 3
    lm.search(transactions.head(8), -1)



from pytest import fixture, mark, raises

from composeml import LabelMaker


@fixture
def data_slice(transactions):
    lm = LabelMaker(
        target_dataframe_index="customer_id", time_index="time", window_size="1h"
    )
    ds = next(lm.slice(transactions, num_examples_per_instance=1))
    return ds


def test_context(data_slice):
    print(data_slice.context)
    context = str(data_slice.context)
    actual = context.splitlines()

    expected = [
        "customer_id                       0",
        "slice_number                      1",
        "slice_start     2019-01-01 08:00:00",
        "slice_stop      2019-01-01 09:00:00",
        "next_start      2019-01-01 09:00:00",
    ]

    assert actual == expected


def test_context_aliases(data_slice):
    assert data_slice.context == data_slice.ctx
    assert data_slice.context.slice_number == data_slice.ctx.count
    assert data_slice.context.slice_start == data_slice.ctx.start
    assert data_slice.context.slice_stop == data_slice.ctx.stop


@mark.parametrize(
    "time_based,offsets",
    argvalues=[
        [False, (2, 4, 2)],
        [False, (2, -6, 2)],
        [True, ("1h", "2h", "1h")],
        [True, ("1h", "-2h30min", "1h")],
        [True, ("2019-01-01 09:00:00", "2019-01-01 10:00:00", "1h")],
    ],
)
def test_subscriptable_slices(transactions, time_based, offsets):
    if time_based:
        dtypes = {"time": "datetime64[ns]"}
        transactions = transactions.astype(dtypes)
        transactions.set_index("time", inplace=True)

    start, stop, size = offsets
    slices = transactions.slice[start:stop:size]
    actual = tuple(map(len, slices))
    assert actual == (2, 2)


def test_subscriptable_error(transactions):
    with raises(TypeError, match="must be a slice object"):
        transactions.slice[0]


def test_time_index_error(transactions):
    match = "offset by frequency requires a time index"
    with raises(AssertionError, match=match):
        transactions.slice[::"1h"]


def test_minimum_data_per_group(transactions):
    lm = LabelMaker(
        "customer_id", labeling_function=len, time_index="time", window_size="1h"
    )
    minimum_data = {1: "2019-01-01 09:00:00", 3: "2019-01-01 12:00:00"}
    lengths = [len(ds) for ds in lm.slice(transactions, 1, minimum_data=minimum_data)]
    assert lengths == [2, 1]


def test_drop_empty(transactions):
    df = transactions.astype({"time": "datetime64[ns]"})
    df.set_index("time", inplace=True)
    df.sort_index(inplace=True)

    ds = df.slice(
        size="1h",
        drop_empty=True,
        stop="2019-01-01 15:00:00",
        start="2019-01-01 08:00:00",
    )

    assert len(list(ds)) == 5



from pytest import raises

from composeml.label_times import LabelTimes
from composeml.tests.utils import to_csv


def test_count_by_time_categorical(total_spent):
    given_answer = total_spent.bin(2, labels=range(2))
    given_answer = to_csv(given_answer.count_by_time)

    answer = [
        "time,0,1",
        "2019-01-01 08:00:00,0,1",
        "2019-01-01 08:30:00,0,2",
        "2019-01-01 09:00:00,0,3",
        "2019-01-01 09:30:00,0,4",
        "2019-01-01 10:00:00,0,5",
        "2019-01-01 10:30:00,1,5",
        "2019-01-01 11:00:00,2,5",
        "2019-01-01 11:30:00,3,5",
        "2019-01-01 12:00:00,4,5",
        "2019-01-01 12:30:00,5,5",
    ]

    assert given_answer == answer


def test_count_by_time_continuous(total_spent):
    given_answer = total_spent.count_by_time
    given_answer = to_csv(given_answer, header=True, index=True)

    answer = [
        "time,total_spent",
        "2019-01-01 08:00:00,1",
        "2019-01-01 08:30:00,2",
        "2019-01-01 09:00:00,3",
        "2019-01-01 09:30:00,4",
        "2019-01-01 10:00:00,5",
        "2019-01-01 10:30:00,6",
        "2019-01-01 11:00:00,7",
        "2019-01-01 11:30:00,8",
        "2019-01-01 12:00:00,9",
        "2019-01-01 12:30:00,10",
    ]

    assert given_answer == answer


def test_sorted_distribution(capsys, total_spent):
    bins = [0, 5, 10, 20]
    total_spent.bin(bins).describe()
    captured = capsys.readouterr()

    out = "\n".join(
        [
            "Label Distribution",
            "------------------",
            "(0, 5]      5",
            "(5, 10]     4",
            "(10, 20]    0",
            "Total:      9",
            "",
            "",
            "Settings",
            "--------",
            "num_examples_per_instance             -1",
            "target_column                total_spent",
            "target_dataframe_index       customer_id",
            "target_type                     discrete",
            "",
            "",
            "Transforms",
            "----------",
            "1. bin",
            "  - bins:         [0, 5, 10, 20]",
            "  - labels:                 None",
            "  - precision:                 3",
            "  - quantiles:             False",
            "  - right:                  True",
            "",
            "",
        ]
    )

    assert captured.out == out


def test_describe_no_transforms(capsys):
    data = {"target": range(3)}
    LabelTimes(data).describe()
    captured = capsys.readouterr()
    out = "\n".join(
        [
            "Label Distribution",
            "------------------",
            "count    3.0",
            "mean     1.0",
            "std      1.0",
            "min      0.0",
            "25%      0.5",
            "50%      1.0",
            "75%      1.5",
            "max      2.0",
            "",
            "",
            "Settings",
            "--------",
            "target_column                 target",
            "target_dataframe_index          None",
            "target_type               continuous",
            "",
            "",
            "Transforms",
            "----------",
            "No transforms applied",
            "",
            "",
        ]
    )

    assert captured.out == out


def test_distribution_categorical(total_spent):
    labels = range(2)
    given_answer = total_spent.bin(2, labels=labels).distribution
    given_answer = to_csv(given_answer)

    answer = [
        "total_spent,count",
        "0,5",
        "1,5",
    ]

    assert given_answer == answer


def test_distribution_continous(total_spent):
    distribution = total_spent.distribution
    actual = to_csv(distribution.round(4))

    expected = [
        ",total_spent",
        "count,10.0",
        "mean,4.5",
        "std,3.0277",
        "min,0.0",
        "25%,2.25",
        "50%,4.5",
        "75%,6.75",
        "max,9.0",
    ]

    assert actual == expected


def test_target_type(total_spent):
    types = total_spent.target_types
    assert types["total_spent"] == "continuous"
    total_spent = total_spent.threshold(5)
    types = total_spent.target_types
    assert types["total_spent"] == "discrete"


def test_count(total_spent):
    given_answer = total_spent.count
    given_answer = to_csv(given_answer, index=True)

    answer = [
        "customer_id,count",
        "0,2",
        "1,3",
        "2,4",
        "3,1",
    ]

    assert given_answer == answer


def test_label_select_errors(total_spent):
    match = "only one target exists"
    with raises(AssertionError, match=match):
        total_spent.select("a")

    lt = total_spent.copy()
    lt.target_columns.append("b")

    match = "target name must be string"
    with raises(TypeError, match=match):
        total_spent.select(123)

    match = 'target "a" not found'
    with raises(AssertionError, match=match):
        lt.select("a")



from pkg_resources import iter_entry_points

for entry_point in iter_entry_points("alteryx_open_src_initialize"):
    try:
        method = entry_point.load()
        if callable(method):
            method("composeml")
    except Exception:
        pass
